"""Main integration back-end."""


import glfw
import OpenGL.GL as gl

import imgui
from imgui.integrations.glfw import GlfwRenderer
from concur.integrations.opengl import create_offscreen_fb, get_fb_data
import time


__pdoc__ = dict(create_window=False, begin_maximized_window=False, create_window_dock=False)


def create_window(window_name, width, height, visible=True):
    """ Create a GLFW window. """
    if not glfw.init():
        print("Could not initialize OpenGL context")
        exit(1)

    # OS X supports only forward-compatible core profiles from 3.2
    glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
    glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
    glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
    glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)
    if not visible:
        glfw.window_hint(glfw.VISIBLE, glfw.FALSE)

    # Create a windowed mode window and its OpenGL context
    window = glfw.create_window(
        int(width), int(height), window_name, None, None
    )
    glfw.make_context_current(window)

    if not window:
        glfw.terminate()
        print("Could not initialize Window")
        exit(1)

    return window


def begin_maximized_window(name, glfw_window):
    imgui.set_next_window_position(0, 0)
    imgui.set_next_window_size(*glfw.get_window_size(glfw_window))
    imgui.push_style_var(imgui.STYLE_WINDOW_ROUNDING, 0)
    imgui.push_style_var(imgui.STYLE_WINDOW_BORDERSIZE, 0)
    window_flags = imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_COLLAPSE | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_BRING_TO_FRONT_ON_FOCUS | imgui.WINDOW_NO_NAV_FOCUS | imgui.WINDOW_NO_DOCKING
    imgui.begin(name, True, window_flags)
    imgui.pop_style_var(2)


def create_window_dock(glfw_window):
    imgui.set_next_window_bg_alpha(0)
    imgui.push_style_var(imgui.STYLE_WINDOW_PADDING, (0, 0))
    begin_maximized_window("Background Window", glfw_window)
    imgui.pop_style_var(1)
    imgui.dock_space("Window Dock Space", 0., 0., 1 << 3)
    imgui.end()


def main(name, widget, width, height, fps=60, save_screencast=None, screencast_fps=60):
    """ Create a GLFW window, spin up the main loop, and display a given widget inside.

    To create a maximized window, pass width and height larger than the screen.

    Arguments:
        widget: The widget to display inside the window. When the widget returns, the application exits.
        width: Desired window width.
        height: Desired window height.
        fps: Maximum number of frames per second
        save_screencast: Capture and save the UI into a specified video file (experimental). Main window shouldn't
            be resized while the application is running when using this option.
        screencast_fps: Save the screencast video with a given FPS.
    """
    imgui.create_context()

    # Set config flags
    imgui.get_io().config_flags |= imgui.CONFIG_DOCKING_ENABLE # | imgui.CONFIG_VIEWPORTS_ENABLE

    window = create_window(name, width, height)
    impl = GlfwRenderer(window)

    ## Using this feels significantly choppier than sleeping manually. TODO: investigate & fix
    # glfw.swap_interval(-1)
    if save_screencast:
        import imageio
        width, height = glfw.get_framebuffer_size(window)
        offscreen_fb = create_offscreen_fb(width, height)
        writer = imageio.get_writer(save_screencast, mode='I', fps=screencast_fps)

    while not glfw.window_should_close(window):
        t0 = time.perf_counter()
        glfw.poll_events()
        impl.process_inputs()

        imgui.new_frame()

        create_window_dock(window)
        begin_maximized_window("Default##Concur", window)

        try:
            next(widget)
        except StopIteration:
            imgui.end()
            imgui.render()
            break
        except:
            # Cleanup on exception for iPython
            imgui.end()
            imgui.render()
            impl.shutdown()
            glfw.terminate()
            if save_screencast:
                writer.close()
            raise

        imgui.end()
        imgui.render()

        if save_screencast:
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, offscreen_fb)
            impl.render(imgui.get_draw_data())
            image = get_fb_data(offscreen_fb, width, height)
            writer.append_data(image)
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)

        impl.render(imgui.get_draw_data())
        glfw.swap_buffers(window)

        t1 = time.perf_counter()
        if t1 - t0 < 1/fps:
            time.sleep(1/fps - (t1 - t0))

    if save_screencast:
        writer.close()

    impl.shutdown()
    glfw.terminate()



"""Pannable, zoomable graph area with axes and gridlines."""

from functools import partial
import imgui
import numpy as np

from concur.extra_widgets.pan_zoom import PanZoom, pan_zoom
import concur.draw as draw
from concur.core import lift, orr, optional, map


margins = [50, 10, -10, -20]


class Frame(PanZoom):
    def __init__(self, top_left, bottom_right, keep_aspect=True, fix_axis=None):
        """
        Simple PanZoom re-export with specified margins.

        Arguments:
            top_left:     Coordinates of the top left corner of the displayed content area.
            bottom_right: Coordinates of the bottom right corner of the displayed content area.
            keep_aspect:  Keep aspect ratio (x/y) equal to a given constant and zoom proportionally.
                          if keep_aspect==True, it is equivalent to keep_aspect==1.
            fix_axis:     Do not zoom in a given axis (`'x'`, `'y'`, or `None`).
        """
        super().__init__(top_left, bottom_right, keep_aspect=keep_aspect, fix_axis=fix_axis, margins=margins)


def _frame(content_gen, show_grid, tf):

    min_tick_spacing=50
    viewport_s = [r + o for r, o in zip(tf.view_s, margins)]
    viewport_c = np.concatenate([np.matmul(tf.s2c, [*viewport_s[:2], 1])[:2], np.matmul(tf.s2c, [*viewport_s[2:], 1])[:2]])
    bg = draw.rect_filled(*tf.view_s, (1,1,1,1))
    if viewport_s[2] <= viewport_s[0] or viewport_s[3] <= viewport_s[1]:
        return bg

    def ticks(a, b, max_n_ticks):
        a, b = min(a, b), max(a, b)
        w = b - a
        min_sep = w / max_n_ticks
        candidates = np.array([10 ** np.floor(np.log10(min_sep)) * f for f in [1, 2, 5, 10]])
        sep = candidates[candidates > min_sep]
        if len(sep) == 0:
            return []
        else:
            sep = sep[0]
            return np.arange(np.ceil(a / sep) * sep, b + 1e-10, sep)

    hticks_c = ticks(viewport_c[3], viewport_c[1], (viewport_s[3] - viewport_s[1]) / min_tick_spacing)
    vticks_c = ticks(viewport_c[2], viewport_c[0], (viewport_s[2] - viewport_s[0]) / min_tick_spacing)
    hticks_s = np.matmul(tf.c2s, np.stack([np.zeros_like(hticks_c), hticks_c, np.ones_like(hticks_c)]))[1]
    vticks_s = np.matmul(tf.c2s, np.stack([vticks_c, np.zeros_like(vticks_c), np.ones_like(vticks_c)]))[0]

    def tick_labels():
        def tick_format(ticks_c, align):
            mag = max(1, abs(ticks_c[0]), abs(ticks_c[-1]))
            stride = abs(ticks_c[1] - ticks_c[0]) if len(ticks_c) > 1 else 1
            some_negative = ticks_c[0] < 0 or ticks_c[-1] < 0
            msd = int(np.floor(np.log10(mag)))
            lsd = int(np.minimum(0, np.floor(np.log10(stride))))
            if 1 + msd - lsd + (lsd < 0) + some_negative > 6:
                format = f"{{:{align}6.1g}}"
            else:
                format = f"{{:{align}6.{-lsd}f}}"
            return format
            print(f"msd: {msd}, stride: {lsd}, n_chars: {1 + msd - lsd + (lsd < 0) + some_negative}")
            format_x = f"{{:<6.{-lsd}f}}"
            format_y = f"{{:>6.{-lsd}f}}"
        xtick_labels = [draw.text(tick_format(vticks_c, "<").format(tc), ts, viewport_s[3], (0,0,0,1)) for ts, tc in zip(vticks_s, vticks_c)]
        ytick_labels = [draw.text(tick_format(hticks_c, ">").format(tc), viewport_s[0] - 45, ts - 7, (0,0,0,1)) for ts, tc in zip(hticks_s, hticks_c)]
        return orr(xtick_labels + ytick_labels)

    def grid():
        hlines = [draw.line(tf.view_s[0], tick, tf.view_s[2], tick, (0,0,0,0.3)) for tick in hticks_s]
        vlines = [draw.line(tick, tf.view_s[1], tick, tf.view_s[3], (0,0,0,0.3)) for tick in vticks_s]
        return orr(hlines + vlines)

    return orr([
        bg,
        lift(imgui.push_clip_rect, *viewport_s, True),
        optional(content_gen is not None, content_gen, tf),
        optional(show_grid, grid),
        lift(imgui.pop_clip_rect),
        draw.rect(*viewport_s, (0,0,0,1)),
        tick_labels(),
    ])


def frame(name, state, content_gen=None, show_grid=True):
    """The frame widget.

    `state` is an instance of `Frame`. See the
    [plot example](https://github.com/potocpav/python-concur/blob/master/examples/plot.py)
    for an usage example.

    Content is specified using `content_gen`, analogously to how it's done in `concur.extra_widgets.image.image`.
    """
    return map(lambda v: ((v[0], v[1][0]) if v[1][0] is not None else v[1][1]),
        pan_zoom(name, state, content_gen=partial(_frame, content_gen, show_grid)))



""" Zoomable, pannable widget with arbitrary content.

This widget is rarely used directly by user code. Most of the time, it's more convenient to use
higher-level widgets `concur.extra_widgets.image.image`, or `concur.extra_widgets.frame.frame`.
"""


import copy
import numpy as np

import imgui
from concur.widgets import child


def pan_zoom(name, state, width=None, height=None, content_gen=None):
    """ Create the Pan & Zoom widget.

    This widget is a pannable, zoomable view of a thing given by `content_gen`. To ease integration with other widgets,
    this widget returns events in a special format: `(name, state, child_event)`. `state` or `child_event` may be `None`
    in case `state` didn't change, or `child_event` didn't fire. `name` is the name supplied to this function.

    `content_gen` is a function that takes the `concur.extra_widgets.pan_zoom.TF` object, and returns a Concur
    widget. It is up to the widget to do the necessary transformations using the `TF` object.
    """
    assert content_gen is not None
    tf, content = None, None
    while True:
        assert not state.keep_aspect or not state.fix_axis, \
            "Can't fix axis and keep_aspect at the same time."
        # Dynamically rescale width and height if they weren't specified
        if width is None:
            w = imgui.get_content_region_available()[0]
        else:
            w = width
        if height is None:
            h = imgui.get_content_region_available()[1]
        else:
            h = height
        frame_w = max(1, w - state.margins[0] + state.margins[2])
        frame_h = max(1, h - state.margins[1] + state.margins[3])
        w = max(1, w)
        h = max(1, h)

        zoom_x = frame_w / (state.right  - state.left)
        zoom_y = frame_h / (state.bottom - state.top)

        left, right = state.left, state.right
        top, bottom = state.top, state.bottom
        if state.keep_aspect:
            aspect = float(state.keep_aspect)
            assert state.keep_aspect > 0, "Negative aspect ratio is not supported."
            if abs(zoom_x) > abs(zoom_y) * aspect:
                zoom_x = np.sign(zoom_x) * abs(zoom_y) * aspect
                center_x = (left + right) / 2
                left  = center_x - frame_w / zoom_x / 2
                right = center_x + frame_w / zoom_x / 2
            if abs(zoom_y) > abs(zoom_x) / aspect:
                zoom_y = np.sign(zoom_y) * abs(zoom_x) / aspect
                center_y = (top + bottom) / 2
                top    = center_y - frame_h / zoom_y / 2
                bottom = center_y + frame_h / zoom_y / 2

        origin = imgui.get_cursor_screen_pos()

        imgui.begin_child("Pan-zoom", w, h, False, flags=imgui.WINDOW_NO_SCROLLBAR | imgui.WINDOW_NO_SCROLL_WITH_MOUSE) # needs to be before is_window_hovered

        # Interaction
        st = copy.deepcopy(state)
        io = imgui.get_io()
        is_mouse_down = io.mouse_down[1] or io.mouse_down[2]
        if is_mouse_down and not st.was_mouse_down and imgui.is_window_hovered():
            st.was_mouse_down = True
            st.is_hovered = imgui.is_window_hovered()
        if not is_mouse_down:
            st.was_mouse_down = False
            st.is_hovered = False

        delta = io.mouse_delta

        # Pan
        if st.is_hovered and (delta[0] or delta[1]):
            if st.fix_axis != 'x':
                st.left -= delta[0] / zoom_x
                st.right -= delta[0] / zoom_x
            if st.fix_axis != 'y':
                st.top -= delta[1] / zoom_y
                st.bottom -= delta[1] / zoom_y

        # Zoom
        if (imgui.is_window_hovered() or st.is_hovered) and io.mouse_wheel:
            factor = 1.3 ** io.mouse_wheel

            if st.fix_axis != 'x':
                mx_rel = (io.mouse_pos[0] - origin[0] - state.margins[0]) / frame_w * 2 - 1
                mx = mx_rel * (right - left) / (st.right - st.left) / 2 + 0.5
                wi = st.right - st.left
                st.left  = st.left    + wi * mx     - wi / factor * mx
                st.right = st.right   - wi * (1-mx) + wi / factor * (1-mx)

            if st.fix_axis != 'y':
                my_rel = (io.mouse_pos[1] - origin[1] - state.margins[1]) / frame_h * 2 - 1
                my = my_rel * (bottom - top) / (st.bottom - st.top) / 2 + 0.5
                hi = st.bottom - st.top
                st.top    = st.top    + hi * my     - hi / factor * my
                st.bottom = st.bottom - hi * (1-my) + hi / factor * (1-my)

        # Get screen-space bounds disregarding the margins
        left_s = left - st.margins[0] / zoom_x
        top_s =   top - st.margins[1] / zoom_y
        right_s = right - st.margins[2] / zoom_x
        bottom_s = bottom - st.margins[3] / zoom_y


        s2c = np.array([ # Screen to Content
            [1 / zoom_x, 0, left_s - origin[0] / zoom_x],
            [0, 1 / zoom_y, top_s  - origin[1] / zoom_y]])

        c2s = np.array([ # Content to Screen
            [zoom_x, 0, origin[0] - left_s * zoom_x],
            [0, zoom_y, origin[1] - top_s  * zoom_y]])

        view_s = [origin[0], origin[1], origin[0] + w, origin[1] + h]
        view_c = [left_s, top_s, right_s, bottom_s]

        content_value = None
        try:
            new_tf = TF(c2s, s2c, view_c, view_s, imgui.is_window_hovered())
            if tf is None or tf != new_tf:
                tf = new_tf
                content = content_gen(tf)
            next(content)
        except StopIteration as e:
            content_value = e.value
        finally:
            imgui.end_child()
        changed = st != state
        if changed or content_value is not None:
            return name, (st if changed else None, content_value)
        else:
            yield


class PanZoom(object):
    """ Pan & zoom state. """
    def __init__(self, top_left, bottom_right, keep_aspect=True, fix_axis=None, margins=[0, 0, 0, 0]):
        """
        Arguments:
            top_left:     Coordinates of the top left corner of the displayed content area.
            bottom_right: Coordinates of the bottom right corner of the displayed content area.
            keep_aspect:  Keep aspect ratio (x/y) equal to a given constant and zoom proportionally.
                          if keep_aspect==True, it is equivalent to keep_aspect==1.
            fix_axis:     Do not zoom in a given axis (`'x'`, `'y'`, or `None`).
            margins:      Margins (left, top, right, bottom) of the view area in pixels.
                          If the view area should be inset by 5 px on each side, then use
                          margins=[5,5,-5,-5].
        """
        self.reset_view(top_left, bottom_right)
         # Include cases where cursor is outside the element, but was inside when the drag started.
         # Exclude cases where cursor is inside the element, but was outside when the drag started.
        self.is_hovered = False
        self.was_mouse_down = False
        self.margins = margins

        self.keep_aspect = keep_aspect
        self.fix_axis = fix_axis

    def reset_view(self, top_left=None, bottom_right=None):
        """ Reset view to default values. """
        if top_left is not None:
            self.default_left = top_left[0]
            self.default_top = top_left[1]
        if bottom_right is not None:
            self.default_right = bottom_right[0]
            self.default_bottom = bottom_right[1]
        self.top = self.default_top
        self.bottom = self.default_bottom
        self.left = self.default_left
        self.right = self.default_right

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class TF(object):
    """ Transformation object containing information necessary for converting between screen-space and image-space.

    Screen-space coordinates are in pixels with top-left window corner equal to (0, 0). Image space is
    arbitrary, given by the respective PanZoom widget state. Transformations are expressed as NumPy
    matrices in homogenous coordinates with two rows and three columns (shape: `(2, 3)`).

    For example, a point `[px, py]` can be transformed to screen-space by left multiplication:

    ```python
    q = np.matmul(c2s, [px, py, 1])
    ```

    Mostly, the necessary conversions are performed by overlay widgets in `concur.draw`.
    It may be useful to transform stuff by hand in cases when default behavior is not sufficient,
    such as specifying the line width in image coordinates.

    Attributes:
        c2s:      Content-to-screen transformation matrix.
        s2c:      Screen-to-content transformation matrix.
        view_s:   Screen-space viewport coordinates as a list [left, top, right, bottom].
        view_c:   Image-space viewport coordinates as a list [left, top, right, bottom].
        hovered:  `True` if the image widget is hovered over.
    """
    def __init__(self, c2s, s2c, view_c, view_s, hovered):
        self.c2s = c2s
        self.s2c = s2c
        self.view_c = view_c
        self.view_s = view_s
        self.hovered = hovered

    def __eq__(self, other):
        return \
            np.array_equal(self.c2s, other.c2s) and \
            np.array_equal(self.s2c, other.s2c) and \
            self.view_c == other.view_c and \
            self.view_s == other.view_s and \
            self.hovered == other.hovered and \
            True

    def transform(self, points):
        """Transform a given NumPy array of `n` points of shape `(n, 2)`."""
        return np.tensordot(np.hstack([points, np.ones((points.shape[0], 1))]), self.c2s, (1, 1))



"""Automation back-end with screen capture and programmatical user input.

Normally, this is used with functions in `concur.testing`.
"""

from __future__ import absolute_import

import glfw
import imgui
import time
import OpenGL.GL as gl

# Save as video
from PIL import Image
import imageio
import numpy as np

from concur.integrations.glfw import create_window, create_window_dock, begin_maximized_window
from concur.integrations.opengl import create_offscreen_fb, get_fb_data
from imgui.integrations import compute_fb_scale
from imgui.integrations.opengl import ProgrammablePipelineRenderer


class PuppetRenderer(ProgrammablePipelineRenderer):
    """Renderer for automated testing. User inputs are set programmatically, rather than interactively."""
    def __init__(self, window):
        super(PuppetRenderer, self).__init__()
        self.window = window

        self.io.display_size = glfw.get_framebuffer_size(self.window)

        self._map_keys()
        self._gui_time = None

        self._click = [False] * 3
        self._mouse_buttons = [False] * 3
        self._mouse_pos = 100, 100
        self._mouse_wheel = 0.0

    def _map_keys(self):
        key_map = self.io.key_map

        key_map[imgui.KEY_TAB] = glfw.KEY_TAB
        key_map[imgui.KEY_LEFT_ARROW] = glfw.KEY_LEFT
        key_map[imgui.KEY_RIGHT_ARROW] = glfw.KEY_RIGHT
        key_map[imgui.KEY_UP_ARROW] = glfw.KEY_UP
        key_map[imgui.KEY_DOWN_ARROW] = glfw.KEY_DOWN
        key_map[imgui.KEY_PAGE_UP] = glfw.KEY_PAGE_UP
        key_map[imgui.KEY_PAGE_DOWN] = glfw.KEY_PAGE_DOWN
        key_map[imgui.KEY_HOME] = glfw.KEY_HOME
        key_map[imgui.KEY_END] = glfw.KEY_END
        key_map[imgui.KEY_DELETE] = glfw.KEY_DELETE
        key_map[imgui.KEY_BACKSPACE] = glfw.KEY_BACKSPACE
        key_map[imgui.KEY_ENTER] = glfw.KEY_ENTER
        key_map[imgui.KEY_ESCAPE] = glfw.KEY_ESCAPE
        key_map[imgui.KEY_A] = glfw.KEY_A
        key_map[imgui.KEY_C] = glfw.KEY_C
        key_map[imgui.KEY_V] = glfw.KEY_V
        key_map[imgui.KEY_X] = glfw.KEY_X
        key_map[imgui.KEY_Y] = glfw.KEY_Y
        key_map[imgui.KEY_Z] = glfw.KEY_Z

    def process_inputs(self):
        """Process the virtual user inputs. Called by `main` at the beginning of each frame."""
        io = imgui.get_io()

        window_size = glfw.get_window_size(self.window)
        fb_size = glfw.get_framebuffer_size(self.window)

        io.display_size = window_size
        io.display_fb_scale = compute_fb_scale(window_size, fb_size)
        io.delta_time = 1.0/60

        current_time = glfw.get_time()

        if self._gui_time:
            self.io.delta_time = current_time - self._gui_time
        else:
            self.io.delta_time = 1. / 60.

        self._gui_time = current_time

        for i, b in enumerate(self._click):
            if b:
                io.mouse_down[i] = True
                self._click[i] = False
            else:
                io.mouse_down[i] = False
        for i, b in enumerate(self._mouse_buttons):
            if b:
                io.mouse_down[i] = True

        io.mouse_pos = self._mouse_pos
        io.mouse_wheel = self._mouse_wheel
        self._mouse_wheel = 0.0

    def click(self, button=0):
        """Simulate a mouse button click.

        * 0 .. left button
        * 1 .. right button
        * 2 .. middle button
        """
        self._click[button] = True

    def set_mouse_pos(self, x, y):
        'Set the mouse cursor position to a specified value.'
        self._mouse_pos = x, y

    def scroll_up(self):
        'Scroll the mouse wheel up one click.'
        self._mouse_wheel = 1

    def scroll_dn(self):
        'Scroll the mouse wheel down one click.'
        self._mouse_wheel = -1

    def mouse_dn(self, button=0):
        """Push a specified mouse button."""
        self._mouse_buttons[button] = True

    def mouse_up(self, button=0):
        """Release a specified mouse button."""
        self._mouse_buttons[button] = False

    def key_dn(self, key):
        imgui.get_io().keys_down[key] = True
        exit(0)

    def key_up(self, key):
        imgui.get_io().keys_down[key] = False

    def write_char(self, c):
        assert 0 < c < 0x10000
        imgui.get_io().add_input_character(c)




def main(name, widget_gen, width, height, save_screencast=None, return_sshot=False, headless=False, fps=60):
    """ Create a GLFW window, spin up the main loop, and display a given widget inside.

    The resulting window is not hooked up to the user input. Instead, input is handled
    by a PuppetRenderer instance.

    `widget_gen` takes as an argument a `PuppetRenderer` instance, and returns a widget.
    `fps` optionally limits FPS (if None, FPS is unlimited)
    """
    imgui.create_context()

    # Set config flags
    imgui.get_io().config_flags |= imgui.CONFIG_DOCKING_ENABLE | imgui.CONFIG_VIEWPORTS_ENABLE

    window = create_window(name, width, height, visible=not headless)
    impl = PuppetRenderer(window)
    widget = widget_gen(impl)
    offscreen_fb = create_offscreen_fb(width, height)

    if save_screencast:
        writer = imageio.get_writer(save_screencast, mode='I', fps=60)

    while not glfw.window_should_close(window):
        t0 = time.perf_counter()
        glfw.poll_events()
        impl.process_inputs()

        imgui.new_frame()

        create_window_dock(window)
        begin_maximized_window("Default##Concur", window)

        try:
            next(widget)
        except StopIteration:
            imgui.end()
            imgui.render()
            break
        except:
            # Cleanup on exception for iPython
            imgui.end()
            imgui.render()
            impl.shutdown()
            glfw.terminate()
            raise

        imgui.end()
        imgui.render()

        if save_screencast:
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, offscreen_fb)
            impl.render(imgui.get_draw_data())
            image = get_fb_data(offscreen_fb, width, height)
            writer.append_data(image)

        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)
        impl.render(imgui.get_draw_data())
        glfw.swap_buffers(window)

        t1 = time.perf_counter()
        if fps is not None and t1 - t0 < 1/fps:
            time.sleep(1/fps - (t1 - t0))

    if save_screencast:
        writer.close()

    if return_sshot:
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, offscreen_fb)
        impl.render(imgui.get_draw_data())
        image = get_fb_data(offscreen_fb, width, height)
        ret = image
    else:
        ret = None

    # # retrieve pixels from framebuffer and write to file
    # file_path = "test.png"
    # image = get_fb_data(offscreen_fb, width, height)
    # image.save(file_path)

    impl.shutdown()
    glfw.terminate()
    return ret



""" Add dragging functionality to any Widget.

Useful for e.g. control points in images. Buttons work well as the base widget, but most graphical
widgets can be used, apart from the `concur.draw` ones. """


import imgui


def draggable(name, widget, tag=None):
    """Add draggable functionality to a widget. Emits (dx, dy) when dragged. The underlying widget works as usual.

    Returned values are equal to the mouse position difference from the previous frame.
    """
    io = imgui.get_io()
    while True:
        try:
            imgui.begin_group()
            next(widget)
        except StopIteration as e:
            return e.value
        finally:
            imgui.end_group()
        if imgui.is_item_active() and (io.mouse_delta[0] or io.mouse_delta[1]):
            return tag if tag is not None else name, io.mouse_delta
        yield



""" Integration with the host system: windowing, user input, etc.

Even using PyImGui's integration code, quite a bit of boilerplate is needed just to create a window,
or to display an image. This module provides functions that simplify getting stuff rendered to just
a couple of lines of code.
"""

import concur.integrations.glfw
import concur.integrations.opengl
import concur.integrations.puppet

from .glfw import *
from .opengl import *
from .quick import quick_plot, quick_window, quick_image



""" Routines for automated testing.

The current automation/testing setup is work-in-progress, and the interface may change radically
in future versions. There are some usability issues that I am not entirely happy with.

See the [tests directory](https://github.com/potocpav/python-concur/tree/master/tests) for usage examples.
"""


import os
import imgui
import numpy as np # for floating point ranges
from functools import partial
from concur.integrations.puppet import PuppetRenderer, main
from concur.draw import polyline
from concur.core import orr, optional


__pdoc__ = dict(test=False)

def test(widget_gen, slow=None, draw_cursor=True, width=512, height=512, *args, **argv):
    return main("Automatic Tester",
        lambda puppet_renderer: widget_gen(draw_cursor, Testing(puppet_renderer, slow)),
        width, height,
        *args, **argv)


def test_widget(f):
    """ Function decorator for testing functions.

    Dead simple usage example, which just displays a button for a moment is:

    ```python
    @c.testing.test_widget
    def test_example(tester):
        yield from c.orr([c.button("Test Button"), tester.pause()])

    if __name__ == '__main__':
        test_example()
    ```

    This can be invoked either directly (`python test_example.py`), or using PyTest (`pytest -k test_example`).
    To slow the test down, set the environmental variable `SLOW_TEST=1`:
    ```bash
    SLOW_TEST=1 python test_example.py
    # or
    SLOW_TEST=1 pytest -k test_example
    ```

    The decorated testing function takes a single argument `tester`, which contains a `Testing` class instance.
    This class provides convenient functions for user input automation, wrapping the raw user interaction
    primitives from `concur.integrations.puppet.PuppetRenderer`.
    """
    def widget_gen(draw_cursor, tester):
        io = imgui.get_io()
        io.mouse_draw_cursor = draw_cursor
        yield from f(tester)
    def g(*args, **argv):
        draw_cursor = 'draw_cursor' in argv and argv['draw_cursor']
        return test(widget_gen, *args, **argv)
    return g


def benchmark_widget(f_gen):
    """ Benchmark a widget (experimental).

    See tests/test_draw.py for example usage.

    TODO: improve, document
    """
    f = f_gen()
    def widget(_):
        for _ in range(240):
            next(f)
            yield
    def g(benchmark):
        benchmark.pedantic(main, ("Perf Tester", widget, 512, 512), dict(fps=None), rounds=1)
    return g


class Testing(object):
    """ Must be used in conjunction with the `concur.integrations.puppet` backend.

    To setup all the plumbing effortlessly, use the `test_widget` decorator.

    All the methods in this class are widgets, and they can be composed as usual using
    `concur.core.orr`, `yield from`, and friends.
    """
    def __init__(self, puppet_renderer, slow=None):
        assert isinstance(puppet_renderer, PuppetRenderer)
        self.puppet = puppet_renderer
        if slow is None:
            self.slow = 'SLOW_TEST' in os.environ and os.environ['SLOW_TEST'] == '1'
        else:
            self.slow = slow
        self.marked = {}

    def click(self, button=0):
        "Click a given mouse button."
        self.puppet.mouse_dn(button)
        if self.slow:
            for i in range(10):
                yield
        yield
        self.puppet.mouse_up(button)

    def click_next(self):
        "Click the next widget."
        x, y = imgui.get_cursor_screen_pos()
        yield from self.move_cursor(x+5, y+5)
        yield from self.pause()
        yield from self.click()
        # Give the widget time to react
        yield
        yield

    def mark(self, name, widget):
        """ Display a widget, but mark it with a name so it can be interacted with at a later point
        using methods such as `click_marked`. """
        while True:
            self.marked[name] = imgui.get_cursor_screen_pos()
            try:
                next(widget)
            except StopIteration as e:
                return e.value
            yield

    def click_marked(self, name, x=5, y=5):
        "Click the given `marked` widget. Optionally, specify the click offset `x, y` coords."
        if name not in self.marked:
            raise ValueError(f"Name '{name}' was not previously marked.")
        x0, y0 = self.marked[name]
        yield from self.move_cursor(x0 + x, y0 + y)
        yield from self.pause()
        yield from self.click()
        yield

    def move_cursor(self, x, y):
        "Move cursor to a given position."
        io = imgui.get_io()
        ox, oy = io.mouse_pos
        yield
        if self.slow:
            for f in np.linspace(0, 1, 30):
                self.puppet.set_mouse_pos(x * f + ox * (1-f), y * f + oy * (1-f))
                yield
        else:
            self.puppet.set_mouse_pos(x, y)
            yield

    def scroll_up(self):
        "Scroll up."
        self.puppet.scroll_up()
        yield

    def scroll_dn(self):
        "Scroll down."
        self.puppet.scroll_dn()
        yield

    def mouse_up(self, button=0):
        "Release the given mouse button."
        self.puppet.mouse_up(button)
        yield

    def mouse_dn(self, button=0):
        "Push the given mouse button."
        self.puppet.mouse_dn(button)
        yield

    def write_char(self, ch):
        "Write a given character."
        self.puppet.write_char(ch)
        yield

    def pause(self, nframes=0):
        """Pause for a specified number of frames.

        If `nframes` <= 0, the pause length depends on the
        environment variable `TEST_SLOW`.
        """
        if nframes <= 0:
            if self.slow:
                for _ in range(30):
                    yield
            yield
        else:
            for _ in range(nframes):
                yield



""" Simple plotting functions designed to monitor a process without slowing it down too much (experimental).

This is not a simple task, as threading in Python is slow, and multiprocessing is fiddly

I am not happy with these interfaces yet, and they may change or be removed in the future.
"""

import threading
import queue
import time

from concur.integrations.glfw import main
from concur.core import listen, tag, multi_orr, nothing
from concur.extra_widgets.frame import frame, Frame
from concur.extra_widgets.image import image, Image


def thread(c, q, quit_q, max_fps):
    t0 = 0
    try:
        for x in c:
            t1 = time.perf_counter()
            if t1 - t0 > 1/max_fps:
                time.sleep(0)
                try:
                    q.put_nowait(x)
                except queue.Full:
                    pass
                t0 = t1
    finally:
        quit_q.put(None)

def quick_plot_w(quit_q, q, x0):
    frame_st = Frame((-1, -1), (1, 1))
    content_gen = x0
    while True:
        events = yield from multi_orr([
            tag("Quit", listen(quit_q)),
            frame("Frame", frame_st, content_gen=content_gen),
            tag("Queue", listen(q)),
            ])
        for tag_, value in events:
            if tag_ == "Queue":
                content_gen = value
            elif tag_ == "Frame":
                frame_st = value
            elif tag_ == "Quit":
                return
        yield


def quick_plot(overlay_gen, w, h, max_fps=60):
    q = queue.Queue(1)
    quit_q = queue.Queue(1)
    t = threading.Thread(target=thread, args=(overlay_gen, q, quit_q, max_fps))

    t.start()
    x0 = q.get()
    main("Quick Plot", quick_plot_w(quit_q, q, x0), w, h)


def quick_window_w(quit_q, q, x0):
    st = nothing()
    while True:
        events = yield from multi_orr([
            tag("Quit", listen(quit_q)),
            st,
            tag("Queue", listen(q)),
            ])
        for tag_, value in events:
            if tag_ == "Queue":
                st = value
            elif tag_ == "Frame":
                frame_st = value
            elif tag_ == "Quit":
                return
        yield



def quick_window(widget_gen, w, h, max_fps=60):
    q = queue.Queue(1)
    quit_q = queue.Queue(1)
    t = threading.Thread(target=thread, args=(widget_gen, q, quit_q, max_fps))

    t.start()
    x0 = q.get()
    main("Quick Window", quick_window_w(quit_q, q, x0), w, h)


def quick_image_w(quit_q, q, x0):
    im_st = Image(x0[0]())
    content_gen = x0[1]
    while True:
        events = yield from multi_orr([
            tag("Quit", listen(quit_q)),
            image("Image", im_st, content_gen=content_gen),
            tag("Queue", listen(q)),
            ])
        for tag_, value in events:
            if tag_ == "Queue":
                content_gen = value[1]
                im_st.change_image(value[0]())
            elif tag_ == "Image":
                im_st = value
            elif tag_ == "Quit":
                return
        yield


def quick_image(overlay_gen, w, h, max_fps=60):
    q = queue.Queue(1)
    quit_q = queue.Queue(1)
    t = threading.Thread(target=thread, args=(overlay_gen, q, quit_q, max_fps))

    t.start()
    x0 = q.get()
    main("Quick Plot", quick_image_w(quit_q, q, x0), w, h)



""" A collection of widgets based off of ImGui.

By convention, all widgets take a string identifier as the first parameter, and return the `(identifier, value)` pair.
This is done for convenience, as Python syntax for mapping values is a bit busy.

Most widgets are just thin wrappers over [widgets in PyImGui](https://pyimgui.readthedocs.io/en/latest/reference/imgui.core.html).
If any widget from ImGui is not listed here, it is mostly trivial to add it. New widgets can be created by writing
generators by hand, or by using helper functions `concur.core.interactive_elem` or `concur.core.lift`.

Some widgets (`transform`, `key_pressed`, etc.) don't do any drawing and serve purely for control flow, or user interaction.

All of the functions in this module are re-exported in the root module for convenience.
"""

__pdoc__ = dict(same_line=False)


import numpy as np # for `transform`
from typing import Iterable, Any, Tuple
import imgui

from concur.core import orr, lift, Widget, interactive_elem


def orr_same_line(widgets):
    """ Use instead of `concur.core.orr` to layout child widgets horizontally instead of vertically. """
    def intersperse(delimiter, seq):
        """ https://stackoverflow.com/questions/5655708/python-most-elegant-way-to-intersperse-a-list-with-an-element """
        from itertools import chain, repeat
        return list(chain.from_iterable(zip(repeat(delimiter), seq)))[1:]

    def group(widget):
        return orr([lift(imgui.begin_group), widget, lift(imgui.end_group)])

    return orr(intersperse(same_line(), [group(w) for w in widgets]))


def window(title: str,
           widget: Widget,
           position: Tuple[int, int] = None,
           size: Tuple[int, int] = None,
           flags: int = 0) -> Widget:
    """ Create a window with a given `widget` inside.

    Window title must be unique. Contents are drawn only if the window is opened.
    """
    while True:
        if position is not None:
            imgui.set_next_window_position(*position)
        if size is not None:
            imgui.set_next_window_size(*size)
        expanded, opened = imgui.begin(title, flags=flags)
        try:
            if expanded and opened:
                next(widget)
        except StopIteration as e:
            return e.value
        finally:
            imgui.end()
        yield


def child(name, widget, width, height, border=False, flags=0):
    """ Create a sized box with a `widget` inside a window. """
    while True:
        imgui.begin_child(name, width, height, border, flags)
        try:
            next(widget)
        except StopIteration as e:
            return e.value
        finally:
            imgui.end_child()
        yield


def collapsing_header(text, widget, open=True):
    """Display a collapsible section header. It can be open or closed by default (parameter `open`).
    """
    while True:
        expanded, visible = imgui.collapsing_header(text, flags=open and imgui.TREE_NODE_DEFAULT_OPEN)
        try:
            if expanded:
                next(widget)
        except StopIteration as e:
            return e.value
        yield


def button(label, tag=None):
    """ Button. Returns `(label, None)` on click, or `(tag, None)` if tag is specified. """
    while not imgui.button(label):
        yield
    return tag if tag is not None else label, None


def color_button(label, color, tag=None):
    """ Colored button. Color is specified as a tuple (R,G,B,A), each in range 0..1.

    Returns `(label, None)` on click, or `(tag, None)` if tag is specified.
    """
    r, g, b, a = color
    while not imgui.color_button(label, r, g, b, a):
        yield
    return tag if tag is not None else label, None


def radio_button(label, active, tag=None):
    """ Radio button. Returns `(label, None)` on click, or `(tag, None)` if tag is specified. """
    while not imgui.radio_button(label, active):
        yield
    return (tag if tag is not None else label), None


def input_text(name, value, buffer_length=255, tag=None):
    """ Text input. """
    while True:
        changed, new_value = imgui.input_text(name, value, buffer_length)
        if changed:
            return (name if tag is None else tag), new_value
        else:
            yield


def key_pressed(name, key_index, repeat=True):
    """ Invisible widget that waits for a given key to be pressed.

    No widget must be active at the time to prevent triggering hotkeys by editing text fields.
    Key codes are specified by the integration layer, e.g. `glfw.KEY_A`, or `glfw.KEY_SPACE`.
    """
    while True:
        if not imgui.is_any_item_active() and imgui.is_key_pressed(key_index, repeat):
            return name, None
        yield


def mouse_click(name, button=0):
    """ Invisible widget that waits for a given mouse button to be clicked (default: LMB).

    No interactive widget must be active to trigger this function.
    Event is returned in the format `(name, (x, y))`, where `(x, y)` are the coordinates
    of the clicked position.
    """
    io = imgui.get_io()
    while True:
        if not imgui.is_any_item_active() and imgui.is_mouse_clicked(button):
            return name, io.mouse_pos
        yield


def text_tooltip(text, widget):
    """Display a text tooltip on hover"""
    while True:
        try:
            imgui.begin_group()
            next(widget)
        except StopIteration as e:
            return e.value
        finally:
            imgui.end_group()
        if imgui.is_item_hovered():
            imgui.set_tooltip(text)
        yield


def tooltip(tooltip_widget, widget):
    """ Display a widget tooltip on hover. May contain arbitrary elements, such as images. """
    while True:
        try:
            imgui.begin_group()
            next(widget)
        except StopIteration as e:
            return e.value
        finally:
            imgui.end_group()

        if imgui.is_item_hovered():
            try:
                imgui.begin_tooltip()
                next(tooltip_widget)
            except StopIteration as e:
                return e.value
            finally:
                imgui.end_tooltip()
        yield


def separator():
    """ Horizontal separator. """
    return lift(imgui.separator)

def spacing():
    """ Extra vertical space. """
    return lift(imgui.spacing)

def same_line():
    """ Call between widgets to layout them horizontally.

    This is Consider using `concur.widgets.orr_same_line` instead as it is more robust.
    """
    return lift(imgui.same_line)


def text(s):
    """ Passive text display widget. """
    return lift(imgui.text, s)

def text_colored(s, r, g, b, a=1.):
    """ Passive colored text display widget. """
    return lift(imgui.text_colored, s, r, g, b, a)

def checkbox(label, checked, *args, **kwargs):
    """ Two-state checkbox. """
    return interactive_elem(imgui.checkbox, label, checked, *args, **kwargs)

def drag_float(label, value, *args, **kwargs):
    """ Float selection widget without a slider. """
    return interactive_elem(imgui.drag_float, label, value, *args, **kwargs)

def drag_float2(label, values, *args, **kwargs):
    """ Float selection widget for selecting two `values`. """
    value0, value1 = values
    return interactive_elem(imgui.drag_float2, label, value0, value1, *args, **kwargs)

def drag_float3(label, values, *args, **kwargs):
    """ Float selection widget for selecting three `values`. """
    value0, value1, value2 = values
    return interactive_elem(imgui.drag_float3, label, value0, value1, value2, *args, **kwargs)

def drag_float4(label, values, *args, **kwargs):
    """ Float selection widget without a slider for selecting four `values`. """
    value0, value1, value2, value3 = values
    return interactive_elem(imgui.drag_float4, label, value0, value1, value2, value3, *args, **kwargs)

def input_float(label, value, *args, **kwargs):
    """ Float input widget. """
    return interactive_elem(imgui.input_float, label, value, *args, **kwargs)

def slider_int(label, value, min_value, max_value, *args, **kwargs):
    """ Int selection slider. """
    return interactive_elem(imgui.slider_int, label, value, min_value, max_value, *args, **kwargs)

def slider_float(label, value, min_value, max_value, *args, **kwargs):
    """ Float selection slider. """
    return interactive_elem(imgui.slider_float, label, value, min_value, max_value, *args, **kwargs)


def test_window():
    """ ImGui test window with a multitude of widgets. """
    return lift(imgui.show_test_window)


def columns(elems, identifier=None, border=True, widths=[]):
    """ Table, using the imgui columns API.

    `elems` is a 2D array of widgets
    `widths` is a optional vector of column widths in pixels. May contain
    None values.
    """
    n_columns = len(elems[0])
    for e in elems:
        assert len(e) == n_columns
    accum = []
    accum.append(lift(imgui.columns, n_columns, identifier, border))
    for i, w in enumerate(widths):
        if w is not None:
            accum.append(lift(imgui.set_column_width, i, w))
    for row in elems:
        for widget in row:
            accum.append(widget)
            accum.append(lift(imgui.next_column))
    accum.append(lift(imgui.columns, 1))
    return orr(accum)


def transform(x, y, widget, tf=None):
    # TODO: move somewhere else
    """ Use `concur.extra_widgets.pan_zoom.TF` and a specified position `x, y` to transform a widget.

    Only widget position will be affected (not scaling), and it will be positioned so that its upper left corner
    is at `[x, y]`.
    """
    old_pos = imgui.get_cursor_screen_pos()
    if tf is not None:
        x, y = np.matmul(tf.c2s, [x, y, 1])
    while True:
        try:
            imgui.set_cursor_screen_pos((x, y))
            next(widget)
        except StopIteration as e:
            return e.value
        finally:
            imgui.set_cursor_screen_pos(old_pos)
        yield



""" Passive geometric shape widgets to be drawn as `concur.extra_widgets.image.image` or `concur.extra_widgets.frame.frame` overlay, or on their own.

All these widgets have the following in common:

* They are passive, so they don't need names. For active overlay, use normal widgets, such as buttons, wrapped inside `concur.widgets.transform`.
* They don't do automatic layout. Instead the exact position is specified by hand.
* Color can be specified in several ways:
    * RGBA tuple with values between 0 and 1. For example, `(0.5, 0.5, 1, 1)` is light blue.
    * RGB tuple with values between 0 and 1. The result is opaque.
    * String specifying a color from the [xkcd color set](https://xkcd.com/color/rgb/), for example, `'red'`.
    * `(str, float)` pair, where the first element specifies color, and the second element specifies alpha.
    * A single `int`, specifying ABGR color. For example, `0xffaa0000` is dark blue.
* `tf` is the `concur.extra_widgets.pan_zoom.TF` object specifying transformations from screen-space to image-space and back.
  If no transformation is supplied, the element is drawn in screen space units.

Theses widgets are not re-exported in the root module, and are normally used as `c.draw.line(...)`, etc.
They can be composed normally using the `concur.core.orr` function.
"""

import numpy as np
import imgui
from concur.colors import color_to_rgba
from concur.core import nothing

__pdoc__ = dict(prepare_polyline_points=False)


def line(x0, y0, x1, y1, color, thickness=1, tf=None):
    """ Line connecting two points. """
    if tf is not None:
        [x0, y0], [x1, y1] = tf.transform(np.array([[x0, y0], [x1, y1]]))
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_line(x0, y0, x1, y1, col, thickness)
        yield


def rect(x0, y0, x1, y1, color, thickness=1, rounding=0, tf=None):
    """ Straight non-filled rectangle specified by its two corners. """
    if tf is not None:
        [x0, y0], [x1, y1] = tf.transform(np.array([[x0, y0], [x1, y1]]))
    # Avoid issues with disappearing lines on very large rectangles
    x0, x1 = np.clip([x0, x1], -8192, 8192)
    y0, y1 = np.clip([y0, y1], -8192, 8192)
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_rect(x0, y0, x1, y1, col, rounding, 15 if rounding else 0, thickness)
        yield


def rects(rects, color, thickness=1, tf=None):
    """ Multiple straight non-filled rectangles specified by their two corners.

    `rects` is a NumPy array of shape `(n, 4)`, where `n` is the number of rectangles.
    """
    if len(rects) == 0:
        while True:
            yield
    if tf is not None:
        rects = tf.transform(rects.reshape(-1, 2)).reshape(rects.shape)
    # Avoid issues with disappearing lines on very large rectangles
    rects = np.clip(rects, -8192, 8192)
    polys = np.empty((len(rects), 4, 2), dtype=rects.dtype)
    polys[:,0] = rects[:,:2]
    polys[:,1,0] = rects[:,0]
    polys[:,1,1] = rects[:,3]
    polys[:,2] = rects[:,2:]
    polys[:,3,0] = rects[:,2]
    polys[:,3,1] = rects[:,1]
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_polylines(polys, col, True, thickness)
        yield


def rect_filled(x0, y0, x1, y1, color, rounding=0, tf=None):
    """ Straight non-filled rectangle specified by its two corners. """
    if tf is not None:
        [x0, y0], [x1, y1] = tf.transform(np.array([[x0, y0], [x1, y1]]))
    # Avoid issues with disappearing lines on very large rectangles
    x0, x1 = np.clip([x0, x1], -8192, 8192)
    y0, y1 = np.clip([y0, y1], -8192, 8192)
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_rect_filled(x0, y0, x1, y1, col, rounding, 15 if rounding else 0)
        yield


def circle(cx, cy, radius, color, thickness=1, num_segments=16, tf=None):
    """ Circle specified by its center and radius. """
    if tf is not None:
        assert np.allclose(np.abs(tf.c2s[0,0]), np.abs(tf.c2s[1,1])), \
            "`tf` must be aspect ratio preserving to draw circles. Use `ellipse` instead, if it isn't the case."
        [cx, cy], radius = np.matmul(tf.c2s, [cx, cy, 1]), radius * tf.c2s[0,0]
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_circle(cx, cy, radius, col, num_segments=num_segments, thickness=thickness)
        yield


def prepare_polyline_points(points, tf):
    if len(points) == 0:
        return []
    if not isinstance(points, np.ndarray):
        points = np.array(points)
    if tf is not None:
        points = tf.transform(points)
    return points


def polyline(points, color, closed=False, thickness=1, tf=None):
    """ Polygonal line or a closed polygon.

    `points` is a list of (x, y) tuples, or a NumPy array of equivalent shape.
    """
    points = prepare_polyline_points(points, tf)
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_polyline(points, col, closed, thickness)
        yield


def polygon(points, color, tf=None):
    """ Filled polygon. Points must form a convex area.

    `points` is a list of (x, y) tuples, or a NumPy array of equivalent shape.
    """
    points = prepare_polyline_points(points, tf)
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_convex_poly_filled(points, col)
        yield


def polylines(points, color, closed=False, thickness=1, tf=None):
    """ Multiple polygonal lines with the same length and parameters.

    Calling this function is more efficient than calling `polyline` multiple times, because all the data is given
    to the C++ back-end in one Python call, and because transformation is vectorized.

    `points` is a NumPy array with shape `(n, m, 2)`, where `n` is the number of polylines, and `m` is the number of points
    in each polyline.
    """
    if tf is not None:
        points = tf.transform(points.reshape(-1, 2)).reshape(points.shape)
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_polylines(points, col, closed, thickness)
        yield


def polygons(points, color, tf=None):
    """ Multiple filled polygons with the same length and color.

    Calling this function is more efficient than calling `polygon` multiple times, because all the data is given
    to the C++ back-end in one Python call, and because transformation is vectorized.

    `points` is a NumPy array with shape `(n, m, 2)`, where `n` is the number of polygons, and `m` is the number of points
    in each polyline.
    """
    if tf is not None:
        points = tf.transform(points.reshape(-1, 2)).reshape(points.shape)
    draw_list = imgui.get_window_draw_list()
    col = color_to_rgba(color)
    while True:
        draw_list.add_convex_polys_filled(points, col)
        yield


def text(string, x, y, color, tf=None):
    """ Text, using the default font and font size.

    This is a raw drawing function. Use `concur.widgets.text` instead if you want a text widget.
    """
    if tf is not None:
        x, y = np.matmul(tf.c2s, [x, y, 1])
    col = color_to_rgba(color)
    while True:
        # Text was hanging the application if too far away
        if -8192 < x < 8192 and -8192 < y < 8192:
            draw_list = imgui.get_window_draw_list()
            draw_list.add_text(x, y, col, string)
        yield


def image(tex_id, w, h, tf):
    """ Draw an image with the given width and height.

    This is a raw drawing function. Use `concur.extra_widgets.image.image` instead if you want an image widget.
    """
    x0, y0, x1, y1 = 0, 0, w, h
    if tf is not None:
        [x0, y0], [x1, y1] = tf.transform(np.array([[x0, y0], [x1, y1]]))
    draw_list = imgui.get_window_draw_list()
    l, t, r, b = tf.view_s
    a_s = tf.view_s[:2]
    b_s = tf.view_s[2:]
    a_i = tf.view_c[0] / w, tf.view_c[1] / h
    b_i = tf.view_c[2] / w, tf.view_c[3] / h
    while True:
        draw_list.add_image(tex_id, tuple(a_s), tuple(b_s), tuple(a_i), tuple(b_i))
        yield


def ellipse(mean, cov, sd, color, thickness=1, num_segments=16, tf=None):
    "Ellipse defined by a mean, covariance matrix, and SD."
    [e1, e2], vs = np.linalg.eig(cov)
    assert e1 > 0 and e2 > 0, "cov must be positive-semidefinite"
    v1, v2 = vs.T
    t = np.linspace(0, np.pi*2, num_segments, endpoint=False).reshape(-1, 1)
    el = v1 * np.sin(t) * np.sqrt(e1) * sd + v2 * np.cos(t) * np.sqrt(e2) * sd
    return polyline(el + mean, color, True, thickness, tf=tf)


def ellipses(means, covs, sd, color, thickness=1, num_segments=16, tf=None):
    """Multiple ellipses defined by a means, covariance matrices, and SD.

    The call is very similar to `ellipse`, but `mean` and `cov` are vectorized: there is one more dimension (zeroth) for both.
    For `n` ellipses, the shape of `mean` is `(n, 2)`, and the shape of `cov` is `(n, 2, 2)`.
    """
    if len(means) == 0 and len(covs) == 0:
        return nothing()
    assert len(means) == len(covs)
    assert len(means.shape) == 2 and means.shape[1] == 2
    assert len(covs.shape) == 3 and covs.shape[1] == 2 and covs.shape[2] == 2
    es, vs = np.linalg.eig(covs)
    assert np.all(es > 0), "covariance matrices must be positive-semidefinite"
    v1 = vs[...,0].reshape(-1, 1, 2)
    v2 = vs[...,1].reshape(-1, 1, 2)
    e1 = es[:,0].reshape(-1, 1, 1)
    e2 = es[:,1].reshape(-1, 1, 1)
    t = np.linspace(0, np.pi*2, num_segments, endpoint=False).reshape(-1, 1)
    el = v1 * np.sin(t) * np.sqrt(e1) * sd + v2 * np.cos(t) * np.sqrt(e2) * sd
    return polylines(el + means.reshape(-1, 1, 2), color, True, thickness, tf=tf)


def scatter(pts, color, marker, marker_size=10, thickness=1, tf=None):
    """Draw a scatter plot with given marker settings.

    If multiple settings are desired (such as two distinct point colors), call
    this function more than once with different parameters.

    Some markers are more performant than others, depending on the amount of
    generated geometry.

    Arguments:
      pts: NumPy array with shape `(n, 2)`, where `n` is the point count.
      color: Color to draw the markers.
      marker: Marker type. See the table below.
      marker_size: Size of the markers in pixels.
      thickness: Line width for non-filled markers
    ------
    marker | description
    ------ | ---
    `"."`  | filled square
    `"x"`  | cross
    `"+"`  | plus sign
    `"o"`  | non-filled circle
    """
    if tf is not None:
        pts = tf.transform(pts)

    if marker == '.':
        r = marker_size / 2
        polys = np.empty((len(pts), 4, 2))
        polys[:, 0, :] = pts + [-r, -r]
        polys[:, 1, :] = pts + [ r, -r]
        polys[:, 2, :] = pts + [ r,  r]
        polys[:, 3, :] = pts + [-r,  r]
        return polygons(polys, color)
    elif marker == '+':
        r = marker_size / 2
        polys = np.empty((len(pts) * 2, 2, 2))
        polys[0::2, 0, :] = pts - [r, 0]
        polys[0::2, 1, :] = pts + [r, 0]
        polys[1::2, 0, :] = pts - [0, r]
        polys[1::2, 1, :] = pts + [0, r]
        return polylines(polys, color, False, thickness)
    elif marker in ['X', 'x', '×']:
        r = marker_size / np.sqrt(8)
        polys = np.empty((len(pts) * 2, 2, 2))
        polys[0::2, 0, :] = pts - [r, r]
        polys[0::2, 1, :] = pts + [r, r]
        polys[1::2, 0, :] = pts - [-r, r]
        polys[1::2, 1, :] = pts + [-r, r]
        return polylines(polys, color, False, thickness)
    elif marker in ['O', 'o']:
        r = marker_size / 2
        n_verts = 7
        t = np.linspace(0, np.pi * 2, n_verts, endpoint=False)
        polys = np.empty((len(pts), n_verts, 2))
        polys[...,0] = np.sin(t) * r
        polys[...,1] = np.cos(t) * r
        polys += pts.reshape(-1, 1, 2)
        return polylines(polys, color, True, thickness)
    else:
        raise ValueError('Invalid marker')



"""Raw OpenGL functions, mostly for creating and deleting textures."""

import numpy as np
from PIL import Image
from OpenGL.GL import *


__pdoc__ = dict(create_offscreen_fb=False, get_fb_data=False)


def texture(arr):
    """ Create a new OpenGL texture and return its texture ID. """
    if len(arr.shape) == 2:
        arr = np.tile(np.expand_dims(arr, 2), (1, 1, 3))

    if len(arr.shape) == 3:
        if arr.shape[2] == 4:
            rgb_mode = GL_RGBA
        elif arr.shape[2] == 3:
            rgb_mode = GL_RGB
        else:
            raise ValueError(f"Only RGB or RGBA channels supported, instead found {arr.shape[2]} color channels.")
    else:
        raise ValueError(f"Shape rank has to be 2 or 3, but it is {len(arr.shape)}")

    texid = int(glGenTextures(1)) # Conversion from np.uint32 to int
    glBindTexture(GL_TEXTURE_2D, texid)
    # If the array isn't normalized, glTexImage2D may leak reference count
    glTexImage2D(GL_TEXTURE_2D, 0, rgb_mode, arr.shape[1], arr.shape[0],
                 0, rgb_mode, GL_UNSIGNED_BYTE, np.ascontiguousarray(arr, 'u1'))
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
    return texid


def rm_textures(tex_ids):
    """ Delete a list of OpenGL textures. """
    glDeleteTextures(tex_ids)


def rm_texture(tex_id):
    """ Delete a single OpenGL texture. """
    glDeleteTextures([tex_id])


def replace_texture(arr, prev_tex_id):
    """ Convenience function: delete the texture `prev_tex_id`, if not `None`, and create a new one,
    returning its texture ID.
    """
    if prev_tex_id is not None:
        rm_texture(prev_tex_id)
    return texture(arr)


def create_offscreen_fb(width, height):
    texture = glGenTextures(1)
    glBindTexture(GL_TEXTURE_2D, texture)
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, None)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

    # create new framebuffer
    offscreen_fb = glGenFramebuffers(1)
    glBindFramebuffer(GL_FRAMEBUFFER, offscreen_fb)
    # attach texture to framebuffer
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture, 0)
    return offscreen_fb


def get_fb_data(offscreen_fb, width, height):
    glBindFramebuffer(GL_FRAMEBUFFER, offscreen_fb)
    pixels = glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE)
    image = Image.frombytes('RGBA', (width, height), pixels)
    image = image.transpose(Image.FLIP_TOP_BOTTOM)
    return np.array(image)



""" Color dictionary. [XKCD colors](https://xkcd.com/color/rgb/) are used. """


import imgui
import sys


MAX_COLOR = 256 - sys.float_info.epsilon * 128


def color_to_rgba(c):
    if isinstance(c, int):
        return c
    elif isinstance(c, tuple) and len(c) == 4:
        return int(MAX_COLOR * c[3]) * 0x1000000 + int(MAX_COLOR * c[2]) * 0x10000 + int(MAX_COLOR * c[1]) * 0x100 + int(MAX_COLOR * c[0])
    elif isinstance(c, tuple) and len(c) == 3:
        return 0xff000000 + int(MAX_COLOR * c[2]) * 0x10000 + int(MAX_COLOR * c[1]) * 0x100 + int(MAX_COLOR * c[0])
    elif isinstance(c, str):
        return xkcd_colors[c]
    elif isinstance(c, tuple) and len(c) == 2 and isinstance(c[0], str):
        return xkcd_colors[c[0]] - 0xff000000 + (min(255, max(0, int(MAX_COLOR * c[1]))) << 24)
    else:
        raise ValueError("Color must be RGBA tuple, RGB tuple, string, int, or (str, float) name & alpha tuple")


# License: http://creativecommons.org/publicdomain/zero/1.0/
xkcd_colors = {
"cloudy blue": 0xffd9c2ac,
"dark pastel green": 0xff57ae56,
"dust": 0xff6e99b2,
"electric lime": 0xff04ffa8,
"fresh green": 0xff4fd869,
"light eggplant": 0xff854589,
"nasty green": 0xff3fb270,
"really light blue": 0xffffffd4,
"tea": 0xff7cab65,
"warm purple": 0xff8f2e95,
"yellowish tan": 0xff81fcfc,
"cement": 0xff91a3a5,
"dark grass green": 0xff048038,
"dusty teal": 0xff85904c,
"grey teal": 0xff8a9b5e,
"macaroni and cheese": 0xff35b4ef,
"pinkish tan": 0xff829bd9,
"spruce": 0xff385f0a,
"strong blue": 0xfff7060c,
"toxic green": 0xff2ade61,
"windows blue": 0xffbf7837,
"blue blue": 0xffc74222,
"blue with a hint of purple": 0xffc63c53,
"booger": 0xff3cb59b,
"bright sea green": 0xffa6ff05,
"dark green blue": 0xff57631f,
"deep turquoise": 0xff747301,
"green teal": 0xff77b50c,
"strong pink": 0xff8907ff,
"bland": 0xff8ba8af,
"deep aqua": 0xff7f7808,
"lavender pink": 0xffd785dd,
"light moss green": 0xff75c8a6,
"light seafoam green": 0xffb5ffa7,
"olive yellow": 0xff09b7c2,
"pig pink": 0xffa58ee7,
"deep lilac": 0xffbd6e96,
"desert": 0xff60adcc,
"dusty lavender": 0xffa886ac,
"purpley grey": 0xff947e94,
"purply": 0xffb23f98,
"candy pink": 0xffe963ff,
"light pastel green": 0xffa5fbb2,
"boring green": 0xff65b363,
"kiwi green": 0xff3fe58e,
"light grey green": 0xffa1e1b7,
"orange pink": 0xff526fff,
"tea green": 0xffa3f8bd,
"very light brown": 0xff83b6d3,
"egg shell": 0xffc4fcff,
"eggplant purple": 0xff410543,
"powder pink": 0xffd0b2ff,
"reddish grey": 0xff707599,
"baby shit brown": 0xff0d90ad,
"liliac": 0xfffd8ec4,
"stormy blue": 0xff9c7b50,
"ugly brown": 0xff03717d,
"custard": 0xff78fdff,
"darkish pink": 0xff7d46da,
"deep brown": 0xff000241,
"greenish beige": 0xff79d1c9,
"manilla": 0xff86faff,
"off blue": 0xffae8456,
"battleship grey": 0xff857c6b,
"browny green": 0xff0a6c6f,
"bruise": 0xff71407e,
"kelley green": 0xff379300,
"sickly yellow": 0xff29e4d0,
"sunny yellow": 0xff17f9ff,
"azul": 0xffec5d1d,
"darkgreen": 0xff074905,
"green/yellow": 0xff08ceb5,
"lichen": 0xff7bb68f,
"light light green": 0xffb0ffc8,
"pale gold": 0xff6cdefd,
"sun yellow": 0xff22dfff,
"tan green": 0xff70bea9,
"burple": 0xffe33268,
"butterscotch": 0xff47b1fd,
"toupe": 0xff7dacc7,
"dark cream": 0xff9af3ff,
"indian red": 0xff040e85,
"light lavendar": 0xfffec0ef,
"poison green": 0xff14fd40,
"baby puke green": 0xff06c4b6,
"bright yellow green": 0xff00ff9d,
"charcoal grey": 0xff42413c,
"squash": 0xff15abf2,
"cinnamon": 0xff064fac,
"light pea green": 0xff82fec4,
"radioactive green": 0xff1ffa2c,
"raw sienna": 0xff00629a,
"baby purple": 0xfff79bca,
"cocoa": 0xff425f87,
"light royal blue": 0xfffe2e3a,
"orangeish": 0xff498dfd,
"rust brown": 0xff03318b,
"sand brown": 0xff60a5cb,
"swamp": 0xff398369,
"tealish green": 0xff73dc0c,
"burnt siena": 0xff0352b7,
"camo": 0xff4e8f7f,
"dusk blue": 0xff8d5326,
"fern": 0xff50a963,
"old rose": 0xff897fc8,
"pale light green": 0xff99fcb1,
"peachy pink": 0xff8a9aff,
"rosy pink": 0xff8e68f6,
"light bluish green": 0xffa8fd76,
"light bright green": 0xff5cfe53,
"light neon green": 0xff54fd4e,
"light seafoam": 0xffbffea0,
"tiffany blue": 0xffdaf27b,
"washed out green": 0xffa6f5bc,
"browny orange": 0xff026bca,
"nice blue": 0xffb07a10,
"sapphire": 0xffab3821,
"greyish teal": 0xff919f71,
"orangey yellow": 0xff15b9fd,
"parchment": 0xffaffcfe,
"straw": 0xff79f6fc,
"very dark brown": 0xff00021d,
"terracota": 0xff4368cb,
"ugly blue": 0xff8a6631,
"clear blue": 0xfffd7a24,
"creme": 0xffb6ffff,
"foam green": 0xffa9fd90,
"grey/green": 0xff7da186,
"light gold": 0xff5cdcfd,
"seafoam blue": 0xffb6d178,
"topaz": 0xffafbb13,
"violet pink": 0xfffc5ffb,
"wintergreen": 0xff86f920,
"yellow tan": 0xff6ee3ff,
"dark fuchsia": 0xff59079d,
"indigo blue": 0xffb1183a,
"light yellowish green": 0xff89ffc2,
"pale magenta": 0xffad67d7,
"rich purple": 0xff580072,
"sunflower yellow": 0xff03daff,
"green/blue": 0xff8dc001,
"leather": 0xff3474ac,
"racing green": 0xff004601,
"vivid purple": 0xfffa0099,
"dark royal blue": 0xff6f0602,
"hazel": 0xff18768e,
"muted pink": 0xff8f76d1,
"booger green": 0xff03b496,
"canary": 0xff63fffd,
"cool grey": 0xffa6a395,
"dark taupe": 0xff4e687f,
"darkish purple": 0xff731975,
"true green": 0xff049408,
"coral pink": 0xff6361ff,
"dark sage": 0xff568559,
"dark slate blue": 0xff614721,
"flat blue": 0xffa8733c,
"mushroom": 0xff889eba,
"rich blue": 0xfff91b02,
"dirty purple": 0xff654a73,
"greenblue": 0xff8bc423,
"icky green": 0xff22ae8f,
"light khaki": 0xffa2f2e6,
"warm blue": 0xffdb574b,
"dark hot pink": 0xff6601d9,
"deep sea blue": 0xff825401,
"carmine": 0xff16029d,
"dark yellow green": 0xff028f72,
"pale peach": 0xffade5ff,
"plum purple": 0xff50054e,
"golden rod": 0xff08bcf9,
"neon red": 0xff3a07ff,
"old pink": 0xff8679c7,
"very pale blue": 0xfffeffd6,
"blood orange": 0xff034bfe,
"grapefruit": 0xff5659fd,
"sand yellow": 0xff66e1fc,
"clay brown": 0xff3d71b2,
"dark blue grey": 0xff4d3b1f,
"flat green": 0xff4c9d69,
"light green blue": 0xffa2fc56,
"warm pink": 0xff8155fb,
"dodger blue": 0xfffc823e,
"gross green": 0xff16bfa0,
"ice": 0xfffaffd6,
"metallic blue": 0xff8e734f,
"pale salmon": 0xff9ab1ff,
"sap green": 0xff158b5c,
"algae": 0xff68ac54,
"bluey grey": 0xffb0a089,
"greeny grey": 0xff7aa07e,
"highlighter green": 0xff06fc1b,
"light light blue": 0xfffbffca,
"light mint": 0xffbbffb6,
"raw umber": 0xff095ea7,
"vivid blue": 0xffff2e15,
"deep lavender": 0xffb75e8d,
"dull teal": 0xff8f9e5f,
"light greenish blue": 0xffb4f763,
"mud green": 0xff026660,
"pinky": 0xffaa86fc,
"red wine": 0xff34008c,
"shit green": 0xff008075,
"tan brown": 0xff4c7eab,
"darkblue": 0xff640703,
"rosa": 0xffa486fe,
"lipstick": 0xff4e17d5,
"pale mauve": 0xfffcd0fe,
"claret": 0xff180068,
"dandelion": 0xff08dffe,
"orangered": 0xff0f42fe,
"poop green": 0xff007c6f,
"ruby": 0xff4701ca,
"dark": 0xff31241b,
"greenish turquoise": 0xffb0fb00,
"pastel red": 0xff5658db,
"piss yellow": 0xff18d6dd,
"bright cyan": 0xfffefd41,
"dark coral": 0xff4e52cf,
"algae green": 0xff6fc321,
"darkish red": 0xff0803a9,
"reddy brown": 0xff05106e,
"blush pink": 0xff8c82fe,
"camouflage green": 0xff13614b,
"lawn green": 0xff09a44d,
"putty": 0xff8aaebe,
"vibrant blue": 0xfff83903,
"dark sand": 0xff598fa8,
"purple/blue": 0xffd0215d,
"saffron": 0xff09b2fe,
"twilight": 0xff8b514e,
"warm brown": 0xff024e96,
"bluegrey": 0xffb2a385,
"bubble gum pink": 0xffaf69ff,
"duck egg blue": 0xfff4fbc3,
"greenish cyan": 0xffb7fe2a,
"petrol": 0xff6a5f00,
"royal": 0xff93170c,
"butter": 0xff81ffff,
"dusty orange": 0xff3a83f0,
"off yellow": 0xff3ff3f1,
"pale olive green": 0xff7bd2b1,
"orangish": 0xff4a82fc,
"leaf": 0xff34aa71,
"light blue grey": 0xffe2c9b7,
"dried blood": 0xff01014b,
"lightish purple": 0xffe652a5,
"rusty red": 0xff0d2faf,
"lavender blue": 0xfff8888b,
"light grass green": 0xff64f79a,
"light mint green": 0xffb2fba6,
"sunflower": 0xff12c5ff,
"velvet": 0xff510875,
"brick orange": 0xff094ac1,
"lightish red": 0xff4a2ffe,
"pure blue": 0xffe20302,
"twilight blue": 0xff7a430a,
"violet red": 0xff5500a5,
"yellowy brown": 0xff0c8bae,
"carnation": 0xff8f79fd,
"muddy yellow": 0xff05acbf,
"dark seafoam green": 0xff76af3e,
"deep rose": 0xff6747c7,
"dusty red": 0xff4e48b9,
"grey/blue": 0xff8e7d64,
"lemon lime": 0xff28febf,
"purple/pink": 0xffde25d7,
"brown yellow": 0xff0597b2,
"purple brown": 0xff3f3a67,
"wisteria": 0xffc27da8,
"banana yellow": 0xff4bfefa,
"lipstick red": 0xff2f02c0,
"water blue": 0xffcc870e,
"brown grey": 0xff68848d,
"vibrant purple": 0xffde03ad,
"baby green": 0xff9eff8c,
"barf green": 0xff02ac94,
"eggshell blue": 0xfff7ffc4,
"sandy yellow": 0xff73eefd,
"cool green": 0xff64b833,
"pale": 0xffd0f9ff,
"blue/grey": 0xffa38d75,
"hot magenta": 0xffc904f5,
"greyblue": 0xffb5a177,
"purpley": 0xffe45687,
"baby shit green": 0xff179788,
"brownish pink": 0xff797ec2,
"dark aquamarine": 0xff717301,
"diarrhea": 0xff03839f,
"light mustard": 0xff60d5f7,
"pale sky blue": 0xfffef6bd,
"turtle green": 0xff4fb875,
"bright olive": 0xff04bb9c,
"dark grey blue": 0xff5b4629,
"greeny brown": 0xff066069,
"lemon green": 0xff02f8ad,
"light periwinkle": 0xfffcc6c1,
"seaweed green": 0xff6bad35,
"sunshine yellow": 0xff37fdff,
"ugly purple": 0xffa042a4,
"medium pink": 0xff9661f3,
"puke brown": 0xff067794,
"very light pink": 0xfff2f4ff,
"viridian": 0xff67911e,
"bile": 0xff06c3b5,
"faded yellow": 0xff7ffffe,
"very pale green": 0xffbcfdcf,
"vibrant green": 0xff08dd0a,
"bright lime": 0xff05fd87,
"spearmint": 0xff76f81e,
"light aquamarine": 0xffc7fd7b,
"light sage": 0xffacecbc,
"yellowgreen": 0xff0ff9bb,
"baby poo": 0xff0490ab,
"dark seafoam": 0xff7ab51f,
"deep teal": 0xff5a5500,
"heather": 0xffac84a4,
"rust orange": 0xff0855c4,
"dirty blue": 0xff9d823f,
"fern green": 0xff448d54,
"bright lilac": 0xfffb5ec9,
"weird green": 0xff7fe53a,
"peacock blue": 0xff956701,
"avocado green": 0xff22a987,
"faded orange": 0xff4d94f0,
"grape purple": 0xff51145d,
"hot green": 0xff29ff25,
"lime yellow": 0xff1dfed0,
"mango": 0xff2ba6ff,
"shamrock": 0xff4cb401,
"bubblegum": 0xffb56cff,
"purplish brown": 0xff47426b,
"vomit yellow": 0xff0cc1c7,
"pale cyan": 0xfffaffb7,
"key lime": 0xff6effae,
"tomato red": 0xff012dec,
"lightgreen": 0xff7bff76,
"merlot": 0xff390073,
"night blue": 0xff480304,
"purpleish pink": 0xffc84edf,
"apple": 0xff3ccb6e,
"baby poop green": 0xff05988f,
"green apple": 0xff1fdc5e,
"heliotrope": 0xfff54fd9,
"yellow/green": 0xff3dfdc8,
"almost black": 0xff0d0d07,
"cool blue": 0xffb88449,
"leafy green": 0xff3bb751,
"mustard brown": 0xff047eac,
"dusk": 0xff81544e,
"dull brown": 0xff4b6e87,
"frog green": 0xff08bc58,
"vivid green": 0xff10ef2f,
"bright light green": 0xff54fe2d,
"fluro green": 0xff02ff0a,
"kiwi": 0xff43ef9c,
"seaweed": 0xff7bd118,
"navy green": 0xff0a5335,
"ultramarine blue": 0xffdb0518,
"iris": 0xffc45862,
"pastel orange": 0xff4f96ff,
"yellowish orange": 0xff0fabff,
"perrywinkle": 0xffe78c8f,
"tealish": 0xffa8bc24,
"dark plum": 0xff2c013f,
"pear": 0xff5ff8cb,
"pinkish orange": 0xff4c72ff,
"midnight purple": 0xff370128,
"light urple": 0xfff66fb3,
"dark mint": 0xff72c048,
"greenish tan": 0xff7acbbc,
"light burgundy": 0xff5b41a8,
"turquoise blue": 0xffc4b106,
"ugly pink": 0xff8475cd,
"sandy": 0xff7adaf1,
"electric pink": 0xff9004ff,
"muted purple": 0xff875b80,
"mid green": 0xff47a750,
"greyish": 0xff95a4a8,
"neon yellow": 0xff04ffcf,
"banana": 0xff7effff,
"carnation pink": 0xffa77fff,
"tomato": 0xff2640ef,
"sea": 0xff92993c,
"muddy brown": 0xff066888,
"turquoise green": 0xff89f404,
"buff": 0xff9ef6fe,
"fawn": 0xff7bafcf,
"muted blue": 0xff9f713b,
"pale rose": 0xffc5c1fd,
"dark mint green": 0xff73c020,
"amethyst": 0xffc05f9b,
"blue/green": 0xff8e9b0f,
"chestnut": 0xff022874,
"sick green": 0xff2cb99d,
"pea": 0xff20bfa4,
"rusty orange": 0xff0959cd,
"stone": 0xff87a5ad,
"rose red": 0xff3c01be,
"pale aqua": 0xffebffb8,
"deep orange": 0xff014ddc,
"earth": 0xff3e65a2,
"mossy green": 0xff278b63,
"grassy green": 0xff039c41,
"pale lime green": 0xff65ffb1,
"light grey blue": 0xffd4bc9d,
"pale grey": 0xfffefdfd,
"asparagus": 0xff56ab77,
"blueberry": 0xff964146,
"purple red": 0xff470199,
"pale lime": 0xff73fdbe,
"greenish teal": 0xff84bf32,
"caramel": 0xff096faf,
"deep magenta": 0xff5c02a0,
"light peach": 0xffb1d8ff,
"milk chocolate": 0xff1e4e7f,
"ocher": 0xff0c9bbf,
"off green": 0xff53a36b,
"purply pink": 0xffe675f0,
"lightblue": 0xfff6c87b,
"dusky blue": 0xff945f47,
"golden": 0xff03bff5,
"light beige": 0xffb6feff,
"butter yellow": 0xff74fdff,
"dusky purple": 0xff7b5b89,
"french blue": 0xffad6b43,
"ugly yellow": 0xff01c1d0,
"greeny yellow": 0xff08f8c6,
"orangish red": 0xff0536f4,
"shamrock green": 0xff4dc102,
"orangish brown": 0xff035fb2,
"tree green": 0xff197e2a,
"deep violet": 0xff480649,
"gunmetal": 0xff676253,
"blue/purple": 0xffef065a,
"cherry": 0xff3402cf,
"sandy brown": 0xff61a6c4,
"warm grey": 0xff848a97,
"dark indigo": 0xff54091f,
"midnight": 0xff2d0103,
"bluey green": 0xff79b12b,
"grey pink": 0xff9b90c3,
"soft purple": 0xffb56fa6,
"blood": 0xff010077,
"brown red": 0xff052b92,
"medium grey": 0xff7c7f7d,
"berry": 0xff4b0f99,
"poo": 0xff03738f,
"purpley pink": 0xffb93cc8,
"light salmon": 0xff93a9fe,
"snot": 0xff0dbbac,
"easter purple": 0xfffe71c0,
"light yellow green": 0xff7ffdcc,
"dark navy blue": 0xff2e0200,
"drab": 0xff448382,
"light rose": 0xffcbc5ff,
"rouge": 0xff3912ab,
"purplish red": 0xff4b05b0,
"slime green": 0xff04cc99,
"baby poop": 0xff007c93,
"irish green": 0xff299501,
"pink/purple": 0xffe71def,
"dark navy": 0xff350400,
"greeny blue": 0xff95b342,
"light plum": 0xff83579d,
"pinkish grey": 0xffa9acc8,
"dirty orange": 0xff0676c8,
"rust red": 0xff0427aa,
"pale lilac": 0xffffcbe4,
"orangey red": 0xff2442fa,
"primary blue": 0xfff90408,
"kermit green": 0xff00b25c,
"brownish purple": 0xff4e4276,
"murky green": 0xff0e7a6c,
"wheat": 0xff7eddfb,
"very dark purple": 0xff34012a,
"bottle green": 0xff054a04,
"watermelon": 0xff5946fd,
"deep sky blue": 0xfff8750d,
"fire engine red": 0xff0200fe,
"yellow ochre": 0xff069dcb,
"pumpkin orange": 0xff077dfb,
"pale olive": 0xff81ccb9,
"light lilac": 0xffffc8ed,
"lightish green": 0xff60e161,
"carolina blue": 0xfffeb88a,
"mulberry": 0xff4e0a92,
"shocking pink": 0xffa202fe,
"auburn": 0xff01309a,
"bright lime green": 0xff08fe65,
"celadon": 0xffb7fdbe,
"pinkish brown": 0xff6172b1,
"poo brown": 0xff015f88,
"bright sky blue": 0xfffecc02,
"celery": 0xff95fdc1,
"dirt brown": 0xff396583,
"strawberry": 0xff4329fb,
"dark lime": 0xff01b784,
"copper": 0xff2563b6,
"medium brown": 0xff12517f,
"muted green": 0xff52a05f,
"robin's egg": 0xfffded6d,
"bright aqua": 0xffeaf90b,
"bright lavender": 0xffff60c7,
"ivory": 0xffcbffff,
"very light purple": 0xfffccef6,
"light navy": 0xff845015,
"pink red": 0xff4f05f5,
"olive brown": 0xff035464,
"poop brown": 0xff01597a,
"mustard green": 0xff04b5a8,
"ocean green": 0xff73993d,
"very dark blue": 0xff330100,
"dusty green": 0xff73a976,
"light navy blue": 0xff885a2e,
"minty green": 0xff7df70b,
"adobe": 0xff486cbd,
"barney": 0xffb81dac,
"jade green": 0xff6aaf2b,
"bright light blue": 0xfffdf726,
"light lime": 0xff6cfdae,
"dark khaki": 0xff558f9b,
"orange yellow": 0xff01adff,
"ocre": 0xff049cc6,
"maize": 0xff54d0f4,
"faded pink": 0xffac9dde,
"british racing green": 0xff0d4805,
"sandstone": 0xff74aec9,
"mud brown": 0xff0f4660,
"light sea green": 0xffb0f698,
"robin egg blue": 0xfffef18a,
"aqua marine": 0xffbbe82e,
"dark sea green": 0xff5d8711,
"soft pink": 0xffc0b0fd,
"orangey brown": 0xff0260b1,
"cherry red": 0xff2a02f7,
"burnt yellow": 0xff09abd5,
"brownish grey": 0xff5f7786,
"camel": 0xff599fc6,
"purplish grey": 0xff7f687a,
"marine": 0xff602e04,
"greyish pink": 0xff948dc8,
"pale turquoise": 0xffd5fba5,
"pastel yellow": 0xff71feff,
"bluey purple": 0xffc74162,
"canary yellow": 0xff40feff,
"faded red": 0xff4e49d3,
"sepia": 0xff2b5e98,
"coffee": 0xff4c81a6,
"bright magenta": 0xffe808ff,
"mocha": 0xff51769d,
"ecru": 0xffcafffe,
"purpleish": 0xff8d5698,
"cranberry": 0xff3a009e,
"darkish green": 0xff377c28,
"brown orange": 0xff0269b9,
"dusky rose": 0xff7368ba,
"melon": 0xff5578ff,
"sickly green": 0xff1cb294,
"silver": 0xffc7c9c5,
"purply blue": 0xffee1a66,
"purpleish blue": 0xffef4061,
"hospital green": 0xffaae59b,
"shit brown": 0xff04587b,
"mid blue": 0xffb36a27,
"amber": 0xff08b3fe,
"easter green": 0xff7efd8c,
"soft blue": 0xffea8864,
"cerulean blue": 0xffee6e05,
"golden brown": 0xff017ab2,
"bright turquoise": 0xfff9fe0f,
"red pink": 0xff552afa,
"red purple": 0xff470782,
"greyish brown": 0xff4f6a7a,
"vermillion": 0xff0c32f4,
"russet": 0xff0539a1,
"steel grey": 0xff8a826f,
"lighter purple": 0xfff45aa5,
"bright violet": 0xfffd0aad,
"prussian blue": 0xff774500,
"slate green": 0xff6d8d65,
"dirty pink": 0xff807bca,
"dark blue green": 0xff495200,
"pine": 0xff345d2b,
"yellowy green": 0xff28f1bf,
"dark gold": 0xff1094b5,
"bluish": 0xffbb7629,
"darkish blue": 0xff824101,
"dull red": 0xff3f3fbb,
"pinky red": 0xff4726fc,
"bronze": 0xff0079a8,
"pale teal": 0xffb2cb82,
"military green": 0xff3e7c66,
"barbie pink": 0xffa546fe,
"bubblegum pink": 0xffcc83fe,
"pea soup green": 0xff17a694,
"dark mustard": 0xff0589a8,
"shit": 0xff005f7f,
"medium purple": 0xffa2439e,
"very dark green": 0xff032e06,
"dirt": 0xff456e8a,
"dusky pink": 0xff8b7acc,
"red violet": 0xff68019e,
"lemon yellow": 0xff38fffd,
"pistachio": 0xff8bfac0,
"dull yellow": 0xff5bdcee,
"dark lime green": 0xff01bd7e,
"denim blue": 0xff925b3b,
"teal blue": 0xff9f8801,
"lightish blue": 0xfffd7a3d,
"purpley blue": 0xffe7345f,
"light indigo": 0xffcf5a6d,
"swamp green": 0xff008574,
"brown green": 0xff116c70,
"dark maroon": 0xff08003c,
"hot purple": 0xfff500cb,
"dark forest green": 0xff042d00,
"faded blue": 0xffbb8c65,
"drab green": 0xff519574,
"light lime green": 0xff66ffb9,
"snot green": 0xff00c19d,
"yellowish": 0xff66eefa,
"light blue green": 0xffb3fb7e,
"bordeaux": 0xff2c007b,
"light mauve": 0xffa192c2,
"ocean": 0xff927b01,
"marigold": 0xff06c0fc,
"muddy green": 0xff327465,
"dull orange": 0xff3b86d8,
"steel": 0xff958573,
"electric purple": 0xffff23aa,
"fluorescent green": 0xff08ff08,
"yellowish brown": 0xff017a9b,
"blush": 0xff8e9ef2,
"soft green": 0xff76c26f,
"bright orange": 0xff005bff,
"lemon": 0xff52fffd,
"purple grey": 0xff856f86,
"acid green": 0xff09fe8f,
"pale lavender": 0xfffecfee,
"violet blue": 0xffc90a51,
"light forest green": 0xff53914f,
"burnt red": 0xff05239f,
"khaki green": 0xff398672,
"cerise": 0xff620cde,
"faded purple": 0xff996e91,
"apricot": 0xff6db1ff,
"dark olive green": 0xff034d3c,
"grey brown": 0xff53707f,
"green grey": 0xff6f9277,
"true blue": 0xffcc0f01,
"pale violet": 0xfffaaece,
"periwinkle blue": 0xfffb998f,
"light sky blue": 0xfffffcc6,
"blurple": 0xffcc3955,
"green brown": 0xff034e54,
"bluegreen": 0xff797a01,
"bright teal": 0xffc6f901,
"brownish yellow": 0xff03b0c9,
"pea soup": 0xff019992,
"forest": 0xff09550b,
"barney purple": 0xff9804a0,
"ultramarine": 0xffb10020,
"purplish": 0xff8c5694,
"puke yellow": 0xff0ebec2,
"bluish grey": 0xff978b74,
"dark periwinkle": 0xffd15f66,
"dark lilac": 0xffa56d9c,
"reddish": 0xff4042c4,
"light maroon": 0xff5748a2,
"dusty purple": 0xff875f82,
"terra cotta": 0xff3b64c9,
"avocado": 0xff34b190,
"marine blue": 0xff6a3801,
"teal green": 0xff6fa325,
"slate grey": 0xff6d6559,
"lighter green": 0xff63fd75,
"electric green": 0xff0dfc21,
"dusty blue": 0xffad865a,
"golden yellow": 0xff15c6fe,
"bright yellow": 0xff01fdff,
"light lavender": 0xfffec5df,
"umber": 0xff0064b2,
"poop": 0xff005e7f,
"dark peach": 0xff5d7ede,
"jungle green": 0xff438204,
"eggshell": 0xffd4ffff,
"denim": 0xff8c633b,
"yellow brown": 0xff0094b7,
"dull purple": 0xff7e5984,
"chocolate brown": 0xff001941,
"wine red": 0xff23037b,
"neon blue": 0xffffd904,
"dirty green": 0xff2c7e66,
"light tan": 0xffaceefb,
"ice blue": 0xfffeffd7,
"cadet blue": 0xff96744e,
"dark mauve": 0xff624c87,
"very light blue": 0xffffffd5,
"grey purple": 0xff8c6d82,
"pastel pink": 0xffcdbaff,
"very light green": 0xffbdffd1,
"dark sky blue": 0xffe48e44,
"evergreen": 0xff2a4705,
"dull pink": 0xff9d86d5,
"aubergine": 0xff34073d,
"mahogany": 0xff00014a,
"reddish orange": 0xff1c48f8,
"deep green": 0xff0f5902,
"vomit green": 0xff03a289,
"purple pink": 0xffd83fe0,
"dusty pink": 0xff948ad5,
"faded green": 0xff74b27b,
"camo green": 0xff256552,
"pinky purple": 0xffbe4cc9,
"pink purple": 0xffda4bdb,
"brownish red": 0xff23369e,
"dark rose": 0xff5d48b5,
"mud": 0xff125c73,
"brownish": 0xff576d9c,
"emerald green": 0xff1e8f02,
"pale brown": 0xff6e91b1,
"dull blue": 0xff9c7549,
"burnt umber": 0xff0e45a0,
"medium green": 0xff48ad39,
"clay": 0xff506ab6,
"light aqua": 0xffdbff8c,
"light olive green": 0xff5cbea4,
"brownish orange": 0xff2377cb,
"dark aqua": 0xff6b6905,
"purplish pink": 0xffae5dce,
"dark salmon": 0xff535ac8,
"greenish grey": 0xff8dae96,
"jade": 0xff74a71f,
"ugly green": 0xff03977a,
"dark beige": 0xff6293ac,
"emerald": 0xff49a001,
"pale red": 0xff4d54d9,
"light magenta": 0xfff75ffa,
"sky": 0xfffcca82,
"light cyan": 0xfffcffac,
"yellow orange": 0xff01b0fc,
"reddish purple": 0xff510991,
"reddish pink": 0xff542cfe,
"orchid": 0xffc475c8,
"dirty yellow": 0xff0ac5cd,
"orange red": 0xff1e41fd,
"deep red": 0xff00029a,
"orange brown": 0xff0064be,
"cobalt blue": 0xffa70a03,
"neon pink": 0xff9a01fe,
"rose pink": 0xff9a87f7,
"greyish purple": 0xff917188,
"raspberry": 0xff4901b0,
"aqua green": 0xff93e112,
"salmon pink": 0xff7c7bfe,
"tangerine": 0xff0894ff,
"brownish green": 0xff096e6a,
"red brown": 0xff162e8b,
"greenish brown": 0xff126169,
"pumpkin": 0xff0177e1,
"pine green": 0xff1e480a,
"charcoal": 0xff373834,
"baby pink": 0xffceb7ff,
"cornflower": 0xfff7796a,
"blue violet": 0xffe9065d,
"chocolate": 0xff021c3d,
"greyish green": 0xff7da682,
"scarlet": 0xff1901be,
"green yellow": 0xff27ffc9,
"dark olive": 0xff023e37,
"sienna": 0xff1e56a9,
"pastel purple": 0xffffa0ca,
"terracotta": 0xff4166ca,
"aqua blue": 0xffe9d802,
"sage green": 0xff78b388,
"blood red": 0xff020098,
"deep pink": 0xff6201cb,
"grass": 0xff2dac5c,
"moss": 0xff589976,
"pastel blue": 0xfffebfa2,
"bluish green": 0xff74a610,
"green blue": 0xff8bb406,
"dark tan": 0xff4a88af,
"greenish blue": 0xff878b0b,
"pale orange": 0xff56a7ff,
"vomit": 0xff15a4a2,
"forrest green": 0xff064415,
"dark lavender": 0xff986785,
"dark violet": 0xff3f0134,
"purple blue": 0xffe92d63,
"dark cyan": 0xff8a880a,
"olive drab": 0xff32766f,
"pinkish": 0xff7e6ad4,
"cobalt": 0xff8f481e,
"neon purple": 0xfffe13bc,
"light turquoise": 0xffccf47e,
"apple green": 0xff26cd76,
"dull green": 0xff62a674,
"wine": 0xff3f0180,
"powder blue": 0xfffcd1b1,
"off white": 0xffe4ffff,
"electric blue": 0xffff5206,
"dark turquoise": 0xff5a5c04,
"blue purple": 0xffce2957,
"azure": 0xfff39a06,
"bright red": 0xff0d00ff,
"pinkish red": 0xff450cf1,
"cornflower blue": 0xffd77051,
"light olive": 0xff69bfac,
"grape": 0xff61346c,
"greyish blue": 0xff9d815e,
"purplish blue": 0xfff91e60,
"yellowish green": 0xff16ddb0,
"greenish yellow": 0xff02fdcd,
"medium blue": 0xffbb6f2c,
"dusty rose": 0xff7a73c0,
"light violet": 0xfffcb4d6,
"midnight blue": 0xff350002,
"bluish purple": 0xffe73b70,
"red orange": 0xff063cfd,
"dark magenta": 0xff560096,
"greenish": 0xff68a340,
"ocean blue": 0xff9c7103,
"coral": 0xff505afc,
"cream": 0xffc2ffff,
"reddish brown": 0xff0a2b7f,
"burnt sienna": 0xff0f4eb0,
"brick": 0xff2336a0,
"sage": 0xff73ae87,
"grey green": 0xff739b78,
"white": 0xffffffff,
"robin's egg blue": 0xfff9ef98,
"moss green": 0xff388b65,
"steel blue": 0xff9a7d5a,
"eggplant": 0xff350838,
"light yellow": 0xff7afeff,
"leaf green": 0xff04a95c,
"light grey": 0xffd6dcd8,
"puke": 0xff02a5a5,
"pinkish purple": 0xffd748d6,
"sea blue": 0xff957404,
"pale purple": 0xffd490b7,
"slate blue": 0xff997c5b,
"blue grey": 0xff8e7c60,
"hunter green": 0xff08400b,
"fuchsia": 0xffd90ded,
"crimson": 0xff0f008c,
"pale yellow": 0xff84ffff,
"ochre": 0xff0590bf,
"mustard yellow": 0xff0abdd2,
"light red": 0xff4c47ff,
"cerulean": 0xffd18504,
"pale pink": 0xffdccfff,
"deep blue": 0xff730204,
"rust": 0xff093ca8,
"light teal": 0xffc1e490,
"slate": 0xff726551,
"goldenrod": 0xff05c2fa,
"dark yellow": 0xff0ab6d5,
"dark grey": 0xff373736,
"army green": 0xff165d4b,
"grey blue": 0xffa48b6b,
"seafoam": 0xffadf980,
"puce": 0xff527ea5,
"spring green": 0xff71f9a9,
"dark orange": 0xff0251c6,
"sand": 0xff76cae2,
"pastel green": 0xff9dffb0,
"mint": 0xffb0fe9f,
"light orange": 0xff48aafd,
"bright pink": 0xffb101fe,
"chartreuse": 0xff0af8c1,
"deep purple": 0xff3f0136,
"dark brown": 0xff021c34,
"taupe": 0xff81a2b9,
"pea green": 0xff12ab8e,
"puke green": 0xff07ae9a,
"kelly green": 0xff2eab02,
"seafoam green": 0xffabf97a,
"blue green": 0xff6d7e13,
"khaki": 0xff62a6aa,
"burgundy": 0xff230061,
"dark teal": 0xff4e4d01,
"brick red": 0xff02148f,
"royal purple": 0xff6e004b,
"plum": 0xff410f58,
"mint green": 0xff9fff8f,
"gold": 0xff0cb4db,
"baby blue": 0xfffecfa2,
"yellow green": 0xff2dfbc0,
"bright purple": 0xfffd03be,
"dark red": 0xff000084,
"pale blue": 0xfffefed0,
"grass green": 0xff0b9b3f,
"navy": 0xff3e1501,
"aquamarine": 0xffb2d804,
"burnt orange": 0xff014ec0,
"neon green": 0xff0cff0c,
"bright blue": 0xfffc6501,
"rose": 0xff7562cf,
"light pink": 0xffdfd1ff,
"mustard": 0xff01b3ce,
"indigo": 0xff820238,
"lime": 0xff32ffaa,
"sea green": 0xffa1fc53,
"periwinkle": 0xfffe828e,
"dark pink": 0xff6b41cb,
"olive green": 0xff047a67,
"peach": 0xff7cb0ff,
"pale green": 0xffb5fdc7,
"light brown": 0xff5081ad,
"hot pink": 0xff8d02ff,
"black": 0xff000000,
"lilac": 0xfffda2ce,
"navy blue": 0xff461100,
"royal blue": 0xffaa0405,
"beige": 0xffa6dae6,
"salmon": 0xff6c79ff,
"olive": 0xff0e756e,
"maroon": 0xff210065,
"bright green": 0xff07ff01,
"dark purple": 0xff3e0635,
"mauve": 0xff8171ae,
"forest green": 0xff0c4706,
"aqua": 0xffc9ea13,
"cyan": 0xffffff00,
"tan": 0xff6fb2d1,
"dark blue": 0xff5b0300,
"lavender": 0xffef9fc7,
"turquoise": 0xffacc206,
"dark green": 0xff003503,
"violet": 0xffea0e9a,
"light purple": 0xfff677bf,
"lime green": 0xff05fe89,
"grey": 0xff919592,
"sky blue": 0xfffdbb75,
"yellow": 0xff14ffff,
"magenta": 0xff7800c2,
"light green": 0xff7bf996,
"orange": 0xff0673f9,
"teal": 0xff869302,
"light blue": 0xfffcd095,
"red": 0xff0000e5,
"brown": 0xff003765,
"pink": 0xffc081ff,
"blue": 0xffdf4303,
"green": 0xff1ab015,
"purple": 0xff9c1e7e,
}



"""
.. include:: ../README.md
   :start-after: <!-- Start docs -->
   :end-before: <!-- End docs -->

.. include:: introduction.md

.. include:: troubleshooting.md

# API

Most of the functions and classes from submodules are re-exported in the
`concur` root module. The most convenient way to access them is by importing `concur` qualified:

```python
import concur as c

# nearly anything can be accessed using `c.anything` now.
```

A notable exception is the `concur.draw` functions: they have a bit different semantics (no layout,
no user interaction), so they are separate and can be accessed using `c.draw.something`.
"""

import concur.extra_widgets
import concur.integrations
import concur.draw
import concur.testing

from .core import *
from .widgets import *
from .extra_widgets import *
from .integrations import main, quick_plot, quick_window, quick_image

import functools


def partial(*args, **argv):
    """Re-export of [`functools.partial`](https://docs.python.org/3.9/library/functools.html#functools.partial)
    for convenience
    """
    return functools.partial(*args, **argv)
